@@ -6,7 +6,9 @@
  * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
  */
 
-#include "config.h"
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -19,9 +21,12 @@
 #include "exportfs.h"
 #include "xio.h"
 #include "mountd.h"
+#include "ha-callout.h"
 
 #include <limits.h> /* PATH_MAX */
 
+extern int reverse_resolve;
+
 /* If new path is a link do not destroy it but place the
  * file where the link points.
  */
@@ -46,21 +51,23 @@ slink_safe_rename(const char * oldpath, const char * newpath)
 }
 
 void
-mountlist_add(nfs_export *exp, const char *path)
+mountlist_add(char *host, const char *path)
 {
 	struct rmtabent	xe;
 	struct rmtabent	*rep;
 	int		lockid;
 	long		pos;
 
-	if ((lockid = xflock(_PATH_RMTAB, "a")) < 0)
+	if ((lockid = xflock(_PATH_RMTABLCK, "a")) < 0)
 		return;
 	setrmtabent("r+");
 	while ((rep = getrmtabent(1, &pos)) != NULL) {
 		if (strcmp (rep->r_client,
-			    exp->m_client->m_hostname) == 0
+			    host) == 0
 		    && strcmp(rep->r_path, path) == 0) {
 			rep->r_count++;
+			/* PRC: do the HA callout: */
+			ha_callout("mount", rep->r_client, rep->r_path, rep->r_count);
 			putrmtabent(rep, &pos);
 			endrmtabent();
 			xfunlock(lockid);
@@ -68,13 +75,15 @@ mountlist_add(nfs_export *exp, const char *path)
 		}
 	}
 	endrmtabent();
-	strncpy(xe.r_client, exp->m_client->m_hostname,
+	strncpy(xe.r_client, host,
 		sizeof (xe.r_client) - 1);
 	xe.r_client [sizeof (xe.r_client) - 1] = '\0';
 	strncpy(xe.r_path, path, sizeof (xe.r_path) - 1);
 	xe.r_path [sizeof (xe.r_path) - 1] = '\0';
 	xe.r_count = 1;
 	if (setrmtabent("a")) {
+		/* PRC: do the HA callout: */
+		ha_callout("mount", xe.r_client, xe.r_path, xe.r_count);
 		putrmtabent(&xe, NULL);
 		endrmtabent();
 	}
@@ -82,15 +91,14 @@ mountlist_add(nfs_export *exp, const char *path)
 }
 
 void
-mountlist_del(nfs_export *exp, const char *path)
+mountlist_del(char *hname, const char *path)
 {
 	struct rmtabent	*rep;
 	FILE		*fp;
-	char		*hname = exp->m_client->m_hostname;
 	int		lockid;
 	int		match;
 
-	if ((lockid = xflock(_PATH_RMTAB, "w")) < 0)
+	if ((lockid = xflock(_PATH_RMTABLCK, "w")) < 0)
 		return;
 	if (!setrmtabent("r")) {
 		xfunlock(lockid);
@@ -104,8 +112,11 @@ mountlist_del(nfs_export *exp, const char *path)
 	while ((rep = getrmtabent(1, NULL)) != NULL) {
 		match = !strcmp (rep->r_client, hname)
 			&& !strcmp(rep->r_path, path);
-		if (match)
+		if (match) {
 			rep->r_count--;
+			/* PRC: do the HA callout: */
+			ha_callout("unmount", rep->r_client, rep->r_path, rep->r_count);
+		}
 		if (!match || rep->r_count)
 			fputrmtabent(fp, rep, NULL);
 	}
@@ -128,7 +139,7 @@ mountlist_del_all(struct sockaddr_in *sin)
 	FILE		*fp;
 	int		lockid;
 
-	if ((lockid = xflock(_PATH_RMTAB, "w")) < 0)
+	if ((lockid = xflock(_PATH_RMTABLCK, "w")) < 0)
 		return;
 	if (!(hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET))) {
 		xlog(L_ERROR, "can't get hostname of %s", inet_ntoa(addr));
@@ -151,10 +162,8 @@ mountlist_del_all(struct sockaddr_in *sin)
 	}
 	while ((rep = getrmtabent(1, NULL)) != NULL) {
 		if (strcmp(rep->r_client, hp->h_name) == 0 &&
-		    (exp = auth_authenticate("umountall", sin, rep->r_path))) {
-			export_reset(exp);
+		    (exp = auth_authenticate("umountall", sin, rep->r_path)))
 			continue;
-		}
 		fputrmtabent(fp, rep, NULL);
 	}
 	if (slink_safe_rename(_PATH_RMTABTMP, _PATH_RMTAB) < 0) {
@@ -176,8 +185,10 @@ mountlist_list(void)
 	struct rmtabent		*rep;
 	struct stat		stb;
 	int			lockid;
+	struct in_addr		addr;
+	struct hostent		*he;
 
-	if ((lockid = xflock(_PATH_RMTAB, "r")) < 0)
+	if ((lockid = xflock(_PATH_RMTABLCK, "r")) < 0)
 		return NULL;
 	if (stat(_PATH_RMTAB, &stb) < 0) {
 		xlog(L_ERROR, "can't stat %s", _PATH_RMTAB);
@@ -195,8 +206,15 @@ mountlist_list(void)
 		setrmtabent("r");
 		while ((rep = getrmtabent(1, NULL)) != NULL) {
 			m = (mountlist) xmalloc(sizeof(*m));
-			m->ml_hostname = xstrdup(rep->r_client);
-			m->ml_directory = xstrdup(rep->r_path);
+
+			if (reverse_resolve &&
+			   inet_aton((const char *) rep->r_client, &addr) &&
+			   (he = gethostbyaddr(&addr, sizeof(addr), AF_INET)))
+				m->ml_hostname = xstrdup(he->h_name);
+			else
+				m->ml_hostname = xstrdup(rep->r_client);
+
+ 			m->ml_directory = xstrdup(rep->r_path);
 			m->ml_next = mlist;
 			mlist = m;
 		}
